##-*- Mode: CPerl -*-

## File: Wlanctl::GUI
## Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
## Description:
##  + Gtk/perl WLAN control GUI
##======================================================================

package Wlanctl::GUI;
use Wlanctl::Config;
use Wlanctl qw(:all);
use Linux::Iwtools qw(:all);
use File::Basename;
use Gtk;
use Storable;
use Data::Dumper;
use strict;

our @ISA = qw(Wlanctl);

##======================================================================
## Globals
##======================================================================
our $VERSION = $Wlanctl::VERSION;
our $XPM_DIR = $Wlanctl::Config::XPM_DIR;

##======================================================================
## Object Defaults
##======================================================================
our %DEFAULTS =
  (
   ##-- inherited defaults
   %Wlanctl::DEFAULTS,
  );

our %GUI_DEFAULTS = 
  (
   ##-- Flags
   default_size => [640,360], ##-- default window size
   link_max  => 255, ##-- maximum link quality
   link_nsamples => 10, ##-- number of samples over which to average link quality
   #autoscan_interval => 100, ##-- autoscan-mode polling frequency (ms)
   #autoscan_interval => 1000, ##-- autoscan-mode polling frequency (ms)
   autoscan_interval => 2000, ##-- autoscan-mode polling frequency (ms)
   autoscan => 0,

   ##-- Filename for load/save
   #filename => "$ENV{HOME}/.gtkwlanrc",

   ##-- Fonts
   font_banner => "-*-helvetica-medium-r-normal-*-24-*-*-*-*-*-iso8859-1",
   font_label  => "-*-helvetica-bold-r-normal-*-14-*-*-*-*-*-iso8859-1",
   font_value  => "-*-helvetica-medium-r-normal-*-14-*-*-*-*-*-iso8859-1",

   ##-------- Colors
   color_known => 'black',
   color_unknown => 'red',
   color_border  => 'grey',
   #color_qlabel  => 'blue',
   #color_qlabel  => 'grey',
   color_qlabel  => 'black',

   ##-------- Pixmaps
   xpm_dir => $XPM_DIR,
   ##
   ##-- Pixmaps: Default network (known nets dialog)
   xpm_default  => 'default.xpm',
   ##
   ##-- Pixmaps: logo
   xpm_logo => 'logo.xpm',
   xpm_icon => 'icon.xpm',
   xpm_error => 'error.xpm',
   xpm_question => 'question.xpm',
   ##
   ##-- Pixmaps: Link quality
   xpm_link => 'link.xpm',
   ##
   ##-- Pixmaps: On/Off
   xpm_on  => 'on.xpm',
   xpm_off => 'off.xpm',
   ##
   ##-- Pixmaps: Up/Down
   xpm_up  => 'up.xpm',
   xpm_down => 'down.xpm',
   ##
   ##-- Pixmaps: assoc/disassoc
   xpm_attach => 'attach.xpm',
   xpm_detach => 'detach.xpm',
   xpm_attached => 'attached.xpm',
   xpm_detached => 'detached.xpm',
   xpm_attached_cl => 'attached_cl.xpm',
   xpm_detached_cl => 'detached_cl.xpm',
   ##
   ##-- Pixmaps: scan/noscan
   xpm_scan => 'scan.xpm',
   xpm_noscan=> 'noscan.xpm',
   ##
   ##-- Pixmaps: wep, nowep
   xpm_wep => 'wep.xpm',
   xpm_nowep => 'nowep.xpm',
   xpm_wep_small=> 'wep_small.xpm',
   ##
   ##-- Pixmaps: stop, refresh
   xpm_quit => 'quit.xpm',
   xpm_refresh => 'refresh.xpm',


   ##-- Widgets
   # fonts=>{ banner=>$Gtk::Gdk::Font, ... },
   # styles=>{ banner=>$Gtk::Style, ... },
   # main=>$Gtk::Window, ##-- main window
   #  vbox=>$Gtk::VBox,
   #   menu=>$menu,
   #   toolbar=>$toolbar,
   #   vhbox=>$Gtk::HBox,
   #    vhvbox=>$Gtk::VBox,
   #     clist=>$clist,
   #     infobox=>$infobox,
   #   statusbar=>$statusbar,
  );

##======================================================================
## Constructor
##  + NOTE: before calling Gtk->main, you should call
##    $wg->initialize() by hand (after loading configs, etc.)
##======================================================================

## $wg = Wlanctl::GUI->new(%args)
##  + %args:
##    device => $device, ##-- default: eth1
##    ...
sub new {
  my $that = shift;
  my $wg = $that->SUPER::new(%DEFAULTS,@_);
  $wg->{gui} = {} if (!$wg->{gui});
  %{$wg->{gui}} = (%GUI_DEFAULTS, name=>ref($wg), %{$wg->{gui}});

  return $wg;
}

## undef = $wg->DESTROY()
sub DESTROY {
  my $wg = shift;
  delete($wg->{default_network});
  $wg->_unindex;
}


##======================================================================
## Useful Gtk Utilities
##======================================================================

## ($label,$value) = $wg->label_value_pair($table,$row,$label_text,$value_text,%args)
sub label_value_pair0 {
  my ($wg,$table,$row,$labtxt,$valtxt, %args) = @_;
  $args{labcol} = 0 if (!defined($args{labcol})); ##-- label-column start
  $args{valcol} = 1 if (!defined($args{valcol})); ##-- value column start
  $args{labwd}  = 1 if (!defined($args{labwd}));   ##-- label width (#/columns)
  $args{valwd}  = 1 if (!defined($args{valwd}));   ##-- value width (#/columns)
  $args{xopts}  = [qw(fill expand)] if (!defined($args{xopts})); ##-- attach options
  $args{yopts}  = [qw(fill expand)] if (!defined($args{yopts})); ##-- attach options 
  $args{xpad}   = 5 if (!defined($args{xpad}));
  $args{ypad}   = 5 if (!defined($args{ypad}));
  my @how = (@args{qw(xopts yopts xpad ypad)});

  my $lab = Gtk::Label->new($labtxt);
  $lab->set_style($wg->{gui}{styles}{label});

  ## gtk_misc_set_alignment(label, xalign, yalign)
  ##  + the align parameter is a float [0.0, 1.0], where 0.0 is top/left, 1.0 is
  ##    bottom/right, and 0.5 is centered.
  $lab->set_justify('right');
  $lab->set_alignment(1.0, 0.5);
  $lab->show;
  $table->attach($lab, $args{labcol}, $args{labcol}+$args{labwd}, $row, $row+1, @how);

  my $val = Gtk::Label->new($valtxt);
  $val->set_style($wg->{gui}{styles}{value});
  $val->set_justify('left');
  $val->set_alignment(0.0, 0.5);
  $val->show;
  $table->attach($val, $args{valcol}, $args{valcol}+$args{valwd}, $row, $row+1, @how);

  return ($lab,$val);
};

## ($label,$value) = $wg->label_value_pair($table,$row,$label_text,$value_text,%args)
sub label_value_pair {
  my ($wg,$table,$row,$labtxt,$valtxt, %args) = @_;

  my $val = Gtk::Label->new($valtxt);
  $val->set_style($wg->{gui}{styles}{value});
  $val->set_justify('left');
  $val->set_alignment(0.0, 0.5);

  return $wg->labelled_widget($table,$row,$labtxt,$val,
			      labstyle=>'label',
			      labxopts=>[qw(fill expand)],
			      labyopts=>[qw(fill expand)],
			      %args,
			      );
};


## ($label,$widget) = $wg->labelled_widget($table,$row,$label_text,$value_widget,%args)
##  + packs $label as label on $widget into $table at row $row
sub labelled_widget {
  my ($wg,$table,$row,$labtxt,$val, %args) = @_;
  $args{labstyle} = 'normal' if (!defined($args{labstyle}));
  $args{labcol} = 0 if (!defined($args{labcol})); ##-- label-column start
  $args{valcol} = 1 if (!defined($args{valcol})); ##-- value column start
  $args{labwd}  = 1 if (!defined($args{labwd}));   ##-- label width (#/columns)
  $args{valwd}  = 1 if (!defined($args{valwd}));   ##-- value width (#/columns)

  $args{labxopts}  = [qw(fill)] if (!defined($args{labxopts}));
  $args{labyopts}  = [qw(fill)] if (!defined($args{labyopts}));
  $args{labxpad}   = 5 if (!defined($args{labpad}));
  $args{labypad}   = 5 if (!defined($args{labpad}));

  $args{valxopts}  = [qw(fill expand)] if (!defined($args{valyopts}));
  $args{valyopts}  = [qw(fill expand)] if (!defined($args{valyopts}));
  $args{valxpad}   = 5 if (!defined($args{valxpad}));
  $args{valypad}   = 5 if (!defined($args{valypad}));
  #my @how = (@args{qw(xopts yopts xpad ypad)});

  my $lab = Gtk::Label->new($labtxt);
  $lab->set_style($wg->{gui}{styles}{$args{labstyle}});

  ## gtk_misc_set_alignment(label, xalign, yalign)
  ##  + the align parameter is a float [0.0, 1.0], where 0.0 is top/left, 1.0 is
  ##    bottom/right, and 0.5 is centered.
  $lab->set_justify('right');
  $lab->set_alignment(1.0, 0.5);
  $lab->show;
  $table->attach($lab, $args{labcol}, $args{labcol}+$args{labwd}, $row, $row+1,
		 @args{qw(labxopts labyopts labxpad labypad)});

  $val->show;
  $table->attach($val, $args{valcol}, $args{valcol}+$args{valwd}, $row, $row+1,
		 @args{qw(valxopts valyopts valxpad valypad)});

  return ($lab,$val);
};


##======================================================================
## Generic Gtk Utilities & Black Magic
##======================================================================

## ($gdk_pixmap,$mask) = new_gdk_pixmap($xpm_filename,$window,$transparent_background)
sub new_gdk_pixmap {
  my ($filename,$window,$background) = @_;
  $window->realize if (!$window->realized);
  my $cmap = $window->get_colormap;
  return Gtk::Gdk::Pixmap->colormap_create_from_xpm($window->window,
						    $cmap,
						    $background,
						    $filename);
}

## $gtk_pixmap = new_pixmap($xpm_filename,$window,$background);
sub new_pixmap { return Gtk::Pixmap->new(new_gdk_pixmap(@_)); }


## ($menubar,$item_factory,$accel_group) = get_main_menbar($window, @menu_items)
sub get_main_menubar {
  my ($window,@menu_items) = @_;

  my $menubar;
  my $item_factory;
  my $accel_group;

  $accel_group = new Gtk::AccelGroup();

  # This function initializes the item factory.
  # Param 1: The type of menu - can be 'Gtk::MenuBar', 'Gtk::Menu',
  #          or 'Gtk::OptionMenu'.
  # Param 2: The path of the menu.
  # Param 3: The accelerator group.  The item factory sets up
  #          the accelerator table while generating menus.
  $item_factory = new Gtk::ItemFactory( 'Gtk::MenuBar',
					'<main>',
					$accel_group );

  # This function generates the menu items. Pass the item factory,
  # the number of items in the array, the array itself, and any
  # callback data for the the menu items.
  $item_factory->create_items( @menu_items );

  # Attach the new accelerator group to the window.
  $window->add_accel_group( $accel_group );

  # Finally, return the actual menu bar created by the item factory.
  #*menubar = gtk_item_factory_get_widget (item_factory, "<main>");
  return ( $item_factory->get_widget('<main>'), $item_factory, $accel_group );
}

## undef = destroy_window($caller, $window_ref1, $window_ref2, ...)
##  + from Gtk/Perl examples/test.pl
sub destroy_window {
  my $caller = shift;
  my ($ref);
  foreach $ref (@_) {
    $$ref = undef if (defined($ref) && ref($ref) && defined($$ref));
  }
  0;
}

##======================================================================
## GUI Initialization (must be called by hand)
##======================================================================

## $wg = $wg->initialize()
##  + initializes GUI keys under $wg->{gui}, including:
##     main => $Gtk::Window, ##-- main window
##     ...
sub initialize {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $w = $gui->{main} = Gtk::Window->new();

  ##--------------------------------------
  ## Styles, fonts, etc.
  $wg->create_colors();
  $wg->create_styles();
  $wg->create_pixmaps();
  $wg->create_tooltips();

  $w->window->set_icon(undef, @{$gui->{gdk_pixmaps}{icon}});

  ##--------------------------------------
  ## Main Window
  $w->signal_connect( 'destroy', sub { Gtk->exit( 0 ); } );
  $w->set_title(ref($wg));
  $w->set_default_size(@{$gui->{default_size}});

  ##--------------------------------------
  ## Main box (vertical)
  my $vb = $gui->{vbox} = Gtk::VBox->new(0,0);
  $vb->show;
  $w->add($vb);

  ##--------------------------------------
  ## Menu bar
  $gui->{menu} = $wg->create_menubar();
  $vb->pack_start($gui->{menu}, 0,1,0);

  ##--------------------------------------
  ## Toolbar
  $gui->{toolbar} = $wg->create_toolbar();
  $vb->pack_start($gui->{toolbar}, 0,1,0);

  ##--------------------------------------
  ## Main content: pack logic
  ##-- pack logic
  $gui->{vhbox} = Gtk::HBox->new();
  $gui->{vbox}->pack_start($gui->{vhbox}, 1,1,0);
  $gui->{vhvbox} = Gtk::VBox->new(0,0);
  $gui->{vhbox}->pack_start($gui->{vhvbox},1,1,5);
  $gui->{mainbox} = $gui->{vhvbox};


  ##--------------------------------------
  ## Main content: available network list (clist)
  ##   + hack: sub HBox adds an additional border
  ##
  $gui->{clist} = $wg->create_clist();
  $gui->{mainbox}->pack_start($gui->{clist}{scroller}, 1,1,0);


  ##--------------------------------------
  ## Main content: info box (frame)
  $gui->{infobox} = $wg->create_infobox();
  $gui->{mainbox}->pack_start($gui->{infobox}, 0,1,5);

  ##--------------------------------------
  ## Status bar
  $gui->{statusbar} = $wg->create_statusbar();
  $gui->{mainbox}->pack_start($gui->{statusbar}, 0,1,0);

  ##--------------------------------------
  ## Context menus
  $gui->{ctxmenus} = $wg->create_context_menus();

  ##-- toolbar style
  #$wg->update_toolbar_style('both');

  ##-- all done
  return $wg;
}

##--------------------------------------------------------------
## GUI Initialization: Styles
##--------------------------------------------------------------

## undef = $wg->create_styles()
sub create_styles {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $fonts  = $gui->{fonts} = {};
  my $styles = $gui->{styles} = {};

  ##-- banner
  $fonts->{banner}  = Gtk::Gdk::Font->load($gui->{font_banner});
  $styles->{banner} = Gtk::Style->new;
  $styles->{banner}->font($fonts->{banner});
  #$styles->{canner}->bg('normal', Gtk::Gdk::Color->parse_color('white')); #?

  ##-- label (for <label,value> pairs)
  $fonts->{label}  = Gtk::Gdk::Font->load($gui->{font_label});
  $styles->{label} = Gtk::Style->new;
  $styles->{label}->font($fonts->{label});

  ##-- value (for <label,value> pairs)
  $fonts->{value}  = Gtk::Gdk::Font->load($gui->{font_value});
  $styles->{value} = Gtk::Style->new;
  $styles->{value}->font($fonts->{value});

  ##-- normal
  $styles->{normal} = $gui->{main}->style;
}

##--------------------------------------------------------------
## GUI Initialization: pixmaps
##--------------------------------------------------------------

## undef = $wg->create_pixmaps()
sub create_pixmaps {
  my $wg = shift;
  my $gui  = $wg->{gui};
  my $gpms = $gui->{gdk_pixmaps} = {};

  my $w = $gui->{main};
  $w->realize if (!$w->realized);
  my $bg = $w->style->bg('normal');

  $gpms->{logo} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_logo}", $w,$bg)];
  $gpms->{icon} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_icon}", $w,$bg)];
  $gpms->{error} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_error}", $w,$bg)];
  $gpms->{question} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_question}", $w,$bg)];
  $gpms->{default} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_default}", $w,$bg)];
  $gpms->{link} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_link}", $w,$bg)];
  $gpms->{on}   = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_on}", $w,$bg)];
  $gpms->{off}  = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_off}", $w,$bg)];
  $gpms->{up}   = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_up}", $w,$bg)];
  $gpms->{down} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_down}", $w,$bg)];
  $gpms->{attach} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_attach}", $w,$bg)];
  $gpms->{detach} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_detach}", $w,$bg)];
  $gpms->{attached} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_attached}", $w,$bg)];
  $gpms->{detached} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_detached}", $w,$bg)];
  $gpms->{attached_cl} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_attached_cl}", $w,$bg)];
  $gpms->{detached_cl} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_detached_cl}", $w,$bg)];
  $gpms->{scan} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_scan}", $w,$bg)];
  $gpms->{noscan} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_noscan}", $w,$bg)];
  $gpms->{quit} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_quit}", $w,$bg)];
  $gpms->{refresh} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_refresh}", $w,$bg)];
  $gpms->{wep} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_wep}", $w,$bg)];
  $gpms->{nowep} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_nowep}", $w,$bg)];
  $gpms->{wep_small} = [new_gdk_pixmap("$gui->{xpm_dir}$gui->{xpm_wep_small}", $w,$bg)];

  ##-- create Gtk::Pixmaps
  my $pixmaps = $gui->{pixmaps} = {};
  foreach (keys(%$gpms)) {
    $pixmaps->{$_} = Gtk::Pixmap->new(@{$gpms->{$_}});
  }

  ##-- paranoia: cleanup
  undef $w;
}

##--------------------------------------------------------------
## GUI Initialization: colors
##--------------------------------------------------------------
## undef = $wg->create_colors()
sub create_colors {
  my $wg = shift;
  my $gui  = $wg->{gui};
  my $colors = $gui->{colors} = {};

  my $w = $gui->{main};
  $w->realize if (!$w->realized);

  my $cmap = $gui->{colormap} = $w->window->get_colormap;

  ##-- setup colors
  $colors->{known} = $cmap->color_alloc(Gtk::Gdk::Color->parse_color($gui->{color_known}));

  $colors->{unknown} = $cmap->color_alloc(Gtk::Gdk::Color->parse_color($gui->{color_unknown}));

  $colors->{border} = $cmap->color_alloc(Gtk::Gdk::Color->parse_color($gui->{color_border}));
  $colors->{qlabel} = $cmap->color_alloc(Gtk::Gdk::Color->parse_color($gui->{color_qlabel}));

  ##-- paranoia: cleanup
  undef $w;
}

##--------------------------------------------------------------
## GUI Initialization: tooltips
##--------------------------------------------------------------
sub create_tooltips {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $tips = $gui->{tips} = Gtk::Tooltips->new;
  #my $tipbg = $w->window->get_colormap->color_alloc(Gtk::Gdk::Color->parse_color('lightyellow'));
  my $tipbg = $gui->{main}->style->white;
  my $tipfg = $gui->{main}->style->black;
  $tips->set_colors($tipbg,$tipfg);
  return $tips;
}

##--------------------------------------------------------------
## GUI Initialization: Menu bar
##--------------------------------------------------------------

## $menu = $wg->create_menubar()
sub create_menubar {
  my $wg = shift;
  my $gui = $wg->{gui};

  # This is the GtkItemFactoryEntry structure used to generate new menus.
  # Item 1: The menu path. The letter after the underscore indicates an
  #         accelerator key once the menu is open.
  # Item 2: The accelerator key for the entry
  # Item 3: The callback function.
  # Item 4: The callback action.  This changes the parameters with
  #         which the function is called.  The default is 0.
  # Item 5: The item type, used to define what kind of an item it is.
  #         Here are the possible values:
  #          NULL               -> "<Item>"
  #          ""                 -> "<Item>"
  #          "<Title>"          -> create a title item
  #          "<Item>"           -> create a simple item
  #          "<CheckItem>"      -> create a check item
  #          "<ToggleItem>"     -> create a toggle item
  #          "<RadioItem>"      -> create a radio item
  #          <path>             -> path of a radio item to link against
  #          "<Separator>"      -> create a separator
  #          "<Branch>"         -> create an item to hold sub items (optional)
  #          "<LastBranch>"     -> create a right justified branch 
  my @menu_items =
    (##-- File 
     {
      path        => '/_File/_Open',
      accelerator => '<control>O',
      callback    => sub { $wg->file_open; },
     },
     {
      path        => '/_File/_Save',
      accelerator => '<control>S',
      callback    => sub { $wg->file_save(force=>1); },
     },
     {
      path        => '/_File/Save _As',
      callback    => sub { $wg->file_saveas; },
     },
     {
      path        => '/_File/Save _Default',
      callback    =>
      sub {
	$gui->{filename}="$ENV{HOME}/.gtkwlanrc";
	$wg->file_save();
      },
     },
     { path=>'/File/sep1', type=>'<Separator>', },
     {
      path        => '/_File/_Quit',
      callback    => sub { Gtk->exit( 0 ); },
      accelerator => '<control>Q',
     },

     ##-----------------------------------
     ## Edit
     { path=>'/_Edit', type=>'<Branch>' },
     { path=>'/Edit/_Known Networks', callback=>sub { $wg->edit_known_networks; },
       accelerator  => '<control>K', },
     { path=>'/Edit/Pr_eferences', callback=>sub { $wg->edit_preferences; },
       accelerator  => '<control>E', },


     ##-----------------------------------
     ## Options
     #{path=>'/_Options', type=>'<Branch>' },
     #{
     # path         => '/Options/_Autoscan',
     # type         => '<CheckItem>',
     # callback     => sub { $wg->autoscan_cb(toggle=>1); },
     #},

     ##-----------------------------------
     ## Options/Toolbar-(Text|Icons|Both)
     ##  + hack: we've got to handle active state in the callback...
     #{ path=>'/Options/sep1', type=>'<Separator>' },
     #{ path=>'/Options/Toolbar _Text',  type=>'<RadioItem>',
     #  callback=>sub { $wg->update_toolbar_style('text'); } },
     #{ path=>'/Options/Toolbar _Icons', type=>'<RadioItem>',
     #  callback=>sub { $wg->update_toolbar_style('icons'); } },
     #{ path=>'/Options/Toolbar _Both',  type=>'<RadioItem>',
     #  callback=>sub { $wg->update_toolbar_style('both'); } },

     ##-----------------------------------
     ## Actions
     { path=>'/_Actions', type=> '<Branch>' },
     ##
     ##-- Actions/On,Off
     { path=>'/_Actions/O_n',  callback=>sub { $wg->_enable;  },
       accelerator  => '<control>N', },
     { path=>'/_Actions/O_ff', callback=>sub { $wg->_disable; },
       accelerator  => '<control>F', },
     { path=>'/_Actions/sep1', type=>'<Separator>' },
     ##
     ##-- Actions/Up,Down
     { path=>'/_Actions/_Up',   callback=>sub { $wg->_ifup;   },
       accelerator  => '<control>U', },
     { path=>'/_Actions/Do_wn', callback=>sub { $wg->_ifdown; },
       accelerator  => '<control>W', },
     { path=>'/_Actions/sep1', type=>'<Separator>' },
     ##
     ##-- Actions/Attach,Detach
     { path=>'/_Actions/_Attach', callback=>sub { $wg->_connect;    },
       accelerator  => '<control>A', },
     { path=>'/_Actions/_Detach', callback=>sub { $wg->_disconnect; },
       accelerator  => '<control>D', },
     { path=>'/_Actions/sep1', type=>'<Separator>' },
     ##
     ##-- Actions/Update,Refresh
     { path=>'/Actions/U_pdate', callback=>sub { $wg->update(force=>1); },
       accelerator  => '<control>P', },
     { path=>'/Actions/_Refresh', callback=>sub { $wg->_refresh; },
       accelerator  => '<control>R', },
     {
      path         => '/Actions/Autos_can',
      type         => '<CheckItem>',
      callback     => sub { $wg->autoscan_cb(toggle=>1); },
      accelerator  => '<control>C',
     },

     ##-----------------------------------
     ## Help
     { path=>'/_Help', type=>'<LastBranch>' },
     { path=>'/_Help/_About',
       callback => sub { $wg->display_about; },
       #callback=>$wg->display_about_callback,
     },
    );

  ##-- get the main menubar
  my ($menubar,$ifactory,$agroup) = get_main_menubar($gui->{main},@menu_items);
  $gui->{menubar}        = $menubar;
  $menubar->{menu_items} = [@menu_items];
  $menubar->{factory}    = $ifactory;

  ##-- mess with radio groups: nope
  #my $texti = $ifactory->get_item('/Options/Toolbar Text');
  #my $iconi = $ifactory->get_item('/Options/Toolbar Icons');
  #$iconi->set_group($texti);

  return $menubar;
}

sub update_toolbar_style {
  my ($wg,$caller) = @_;
  my $gui = $wg->{gui};
  my $ifact = $gui->{menubar}{factory};

  if ($caller eq 'text') {
    $gui->{toolbar}->set_style('text');
    $ifact->get_item('/Options/Toolbar Text')->active(1);
    $ifact->get_item('/Options/Toolbar Icons')->active(0);
    $ifact->get_item('/Options/Toolbar Both')->active(0);
  }
  elsif ($caller eq 'icons') {
    $gui->{toolbar}->set_style('icons');
    $ifact->get_item('/Options/Toolbar Text')->active(0);
    $ifact->get_item('/Options/Toolbar Icons')->active(1);
    $ifact->get_item('/Options/Toolbar Both')->active(0);
  }
  else {
    $gui->{toolbar}->set_style('both');
    $ifact->get_item('/Options/Toolbar Text')->active(0);
    $ifact->get_item('/Options/Toolbar Icons')->active(0);
    $ifact->get_item('/Options/Toolbar Both')->active(1);
  }
}

##--------------------------------------------------------------
## GUI Initialization: Toolbar
##--------------------------------------------------------------

## $toolbar = $wg->create_toolbar()
#*create_toolbar = \&create_toolbar1;
sub create_toolbar1 {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $w = $gui->{main};

  $w->realize if (!$w->realized);

  my $bg = $w->style->bg('normal');
  my $toolbar = Gtk::Toolbar->new('horizontal', 'both');
  ##
  ##-- seems not to work at creation time
  #my $toolbar = Gtk::Toolbar->new('horizontal', 'icons');
  #$toolbar->set_style('icons');
  ##
  $toolbar->set_space_style('line');
  $toolbar->set_button_relief('none');
  $toolbar->{text} = $toolbar->{icons} = 1;

  ##-- Buttons: On/Off
  $toolbar->{on} = $toolbar->append_item("On",
					 "Enable WLAN device",
					 "Toolbar/Enable",
					 $gui->{pixmaps}{on}
					);

  $toolbar->{on}->signal_connect('clicked', sub { $wg->enable; });

  $toolbar->{off} = $toolbar->append_item ("Off",
					   "Disable WLAN device",
					   "Toolbar/Disable",
					    $gui->{pixmaps}{off}
					  );
  $toolbar->{off}->signal_connect('clicked', sub { $wg->disable; });

  ##-- Buttons: Up/Down
  #$toolbar->append_space();
  $toolbar->{up} = $toolbar->append_item ("Up",
					  "Connect to selected network",
					  "Toolbar/Up",
					  $gui->{pixmaps}{up}
					 );
  $toolbar->{up}->signal_connect('clicked', sub { $wg->up; });

  $toolbar->{down} = $toolbar->append_item("Down",
					   "Disconnect from current network",
					   "Toolbar/Down",
					   $gui->{pixmaps}{down}
					  );
  $toolbar->{down}->signal_connect('clicked', sub { $wg->down; });

  ##-- Assoc/Dissassoc
  #$toolbar->append_space;
  $toolbar->{attach} = $toolbar->append_item("Attach",
					    "Associate to current network",
					     "Toolbar/Attach",
					     $gui->{pixmaps}{attach}
					    );
  $toolbar->{detach} = $toolbar->append_item("Detach",
					     "Disassociate from current network",
					     "Toolbar/Detach",
					     $gui->{pixmaps}{detach}
					    );

  ##-- autoscan button
  $toolbar->{autoscan} = $toolbar->append_item("Autoscan",
					      "Auto-scan for available networks",
					      "Toolbar/Autoscan",
					      $gui->{pixmaps}{scan}
					     );
  $toolbar->{autoscan}->signal_connect('autoscan', sub { $wg->autoscan_cb(toggle=>1); });

  ##-- refresh button
  #$toolbar->append_space();
  $toolbar->{refresh} = $toolbar->append_item("Refresh",
					      "Update status",
					      "Toolbar/Refresh",
					      $gui->{pixmaps}{refresh}
					     );
  $toolbar->{refresh}->signal_connect('clicked', sub { $wg->update(force=>1); });

  ##-- quit button
  #$toolbar->append_space();
  $toolbar->{quit} = $toolbar->append_item("Exit",
					   "Exit the program",
					   "Toolbar/Exit",
					   $gui->{pixmaps}{quit}
					  );
  $toolbar->{quit}->signal_connect('clicked', sub { Gtk->main_quit } );

  $toolbar->show;
  #$toolbar->set_style('icons'); ##-- not here either

  #my $hbox = $toolbar->{handlebox} = Gtk::HandleBox->new;
  #my $hbox = $toolbar->{handlebox} = Gtk::HBox->new;
  my $hbox = $toolbar->{handlebox} = Gtk::Frame->new;
  $hbox->set_shadow_type('out');
  $hbox->add($toolbar);

  undef $w;
  return $toolbar;
}


## $button = $wg->toolbar_append($label, $tiptext, $tipprivate, $gdk_pixmap_key)
sub toolbar_append {
  my ($wg,$col,$label,$tiptext,$tippriv,$pixkey) = @_;
  my $gui = $wg->{gui};
  my $tbar = $gui->{toolbar};
  my $ttab = $tbar->{table};

  my $button = Gtk::Button->new();
  $button->{vbox} = Gtk::VBox->new(0,0);
  $button->add($button->{vbox});
  $button->set_relief('none');

  $button->{pixmap} = Gtk::Pixmap->new(@{$gui->{gdk_pixmaps}{$pixkey}});
  $button->{pixmap}->show;
  $button->{vbox}->pack_start($button->{pixmap}, 0,0,3);

  $button->{label}  = Gtk::Label->new($label);
  $button->{label}->show;
  $button->{vbox}->pack_start($button->{label}, 0,0,3);

  ##-- packing
  $button->{vbox}->show;
  $button->show;
  my @how = ([qw(fill expand)], [qw(fill expand)], 0,0);
  $ttab->attach($button, $col, $col+1, 0,1, @how);

  ##-- set tooltips
  $button->{tippriv} = $tippriv;
  $gui->{tips}->set_tip($button,$tiptext,$tippriv);

  return $button;
}

## $toolbar = $wg->create_toolbar2();
*create_toolbar = \&create_toolbar2;
sub create_toolbar2 {
  my $wg = shift;
  my $gui = $wg->{gui};

  my $tbar = $gui->{toolbar} = Gtk::Frame->new;
  $tbar->set_shadow_type('out');
  my $ncols = 7;
  my $tbox = $tbar->{box}   = Gtk::HBox->new(0,0);
  $tbar->add($tbox);
  my $ttab = $tbar->{table} = Gtk::Table->new(1,$ncols, 1);
  #$tbar->add($ttab);
  $tbox->pack_start($ttab,0,0,0);

  ##-- On/Off
  my $col=0;
  $tbar->{on} = $wg->toolbar_append($col++, "On",
				    "Enable WLAN device",
				    "Toolbar/On",
				    'on'
				   );
  $tbar->{on}->signal_connect('clicked', sub { $wg->toggle_on; });

  ##-- Up/Down
  $tbar->{up} = $wg->toolbar_append($col++, "Up",
				    "Run up the WLAN device",
				    "Toolbar/Up",
				    'up'
				   );
  $tbar->{up}->signal_connect('clicked', sub { $wg->toggle_up; });

  ##-- Attach/Detach
  $tbar->{attach} = $wg->toolbar_append($col++, "Attach",
					"Attach to currently selected network",
					"Toolbar/Attach",
					'attach'
				       );
  $tbar->{attach}->signal_connect('clicked', sub { $wg->toggle_attach; });

  ##-- Autoscan
  $tbar->{autoscan} = $wg->toolbar_append($col++, "Autoscan",
					  "Auto-scan for available networks",
					  "Toolbar/Autoscan",
					  'scan'
					 );
  $tbar->{autoscan}->signal_connect('clicked', sub { $wg->autoscan_cb(toggle=>1); });


  ##-- Refresh
  $tbar->{refresh} = $wg->toolbar_append($col++, "Refresh",
					"Update status",
					"Toolbar/Refresh",
					'refresh'
				       );
  $tbar->{refresh}->signal_connect('clicked',
				   #sub { $wg->update(force=>1); }
				   sub { $wg->_refresh; },
				  );

  ##-- Quit
  $tbar->{quit} = $wg->toolbar_append($col++, "Quit",
					"Exit the program",
					"Toolbar/Quit",
					'quit'
				       );
  $tbar->{quit}->signal_connect('clicked', sub { Gtk->main_quit });

  return $tbar;
}



##--------------------------------------------------------------
## GUI Initialization: Available network list (clist)
##--------------------------------------------------------------

## $clist = $wg->create_clist()
sub create_clist {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $w = $gui->{main};

  my @titles = ('', 'Name', 'ESSID', 'WEP', 'Link Quality');
  my $clist = Gtk::CList->new_with_titles(@titles);
  $clist->set_column_width(0,24);
  $clist->set_column_width(1,166);
  $clist->set_column_width(2,200);
  #$clist->set_column_width(3,48);
  $clist->set_column_width(4,128);
  $clist->shadow_type('in');
  $clist->column_titles_show;
  $clist->set_selection_mode('single');
  $clist->reorderable(0);
  $clist->set_column_justification(0, 'center');
  $clist->set_column_justification(3, 'center');

  $clist->signal_connect('select_row', sub { $wg->clist_select(@_); });
  $clist->signal_connect('unselect_row', sub { $wg->clist_unselect(@_); });
  #$clist->signal_connect('click_column', sub { print "click_column($_[1])\n"; });
  $clist->signal_connect('button-press-event', sub { $wg->clist_click(@_); });

  ##-- ... make it scrollable
  my $scrolled_win = $clist->{scroller} = Gtk::ScrolledWindow->new(undef, undef);
  $scrolled_win->set_policy('automatic', 'automatic');
  $scrolled_win->add($clist);

  return $clist;
}


##--------------------------------------------------------------
## GUI Initialization: Info box
##--------------------------------------------------------------

# $infobox_frame = $wg->create_infobox()
sub create_infobox {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $w = $gui->{main};

  ##-- add an unlabelled outer frame
  my $hbox = Gtk::HBox->new;

  ##-- add a labelled frame (current network)
  my $cframe = $hbox->{current} = Gtk::Frame->new("Current Network");
  $hbox->pack_start($cframe, 1,1,0);

  my $nrows = 4;
  my $ncols = 2;
  my $tab = $cframe->{table} = Gtk::Table->new($nrows,$ncols);
  #$stab->set_alignment(0,0.5); ##-- nope
  $cframe->add($tab);

  my %args = (ypad=>1);
  $cframe->{name}  = [$wg->label_value_pair($tab, 0, "Name:", "(none)", %args)];
  $cframe->{essid} = [$wg->label_value_pair($tab, 1, "ESSID:", "(none)", %args)];
  $cframe->{ap} = [$wg->label_value_pair($tab, 2, "AP MAC:", "00:00:00:00:00:00", %args)];
  $cframe->{ipaddr} = [$wg->label_value_pair($tab, 3, "IP Addr:", "0.0.0.0", %args)];


  ##-- add a labelled frame (selected network)
  my $sframe = $hbox->{selected} = Gtk::Frame->new("Selected Network");
  $hbox->pack_start($sframe, 1,1,0);

  $nrows = 4;
  $ncols = 2;
  $tab = $sframe->{table} = Gtk::Table->new($nrows,$ncols);
  #$stab->set_alignment(0,0.5); ##-- nope
  $sframe->add($tab);

  $sframe->{name}  = [$wg->label_value_pair($tab, 0, "Name:", "(none)", %args)];
  $sframe->{essid} = [$wg->label_value_pair($tab, 1, "ESSID:", "(none)", %args)];
  $sframe->{ap} = [$wg->label_value_pair($tab, 2, "AP MAC:", "00:00:00:00:00:00", %args)];
  #$sframe->{ipaddr} = [$wg->label_value_pair($tab, 3, "IP Addr:", "0.0.0.0", %args)];

  return $hbox;
}


##--------------------------------------------------------------
## GUI Initialization: Status bar
##--------------------------------------------------------------

## $pixmap = $wg->add_pixmap($pixmap_key, $parent, @packargs)
sub add_statusbar_pixmap {
  my ($wg,$key,$parent,@packargs) = @_;

  #my $pix = Gtk::Pixmap->new(@{$wg->{gui}{gdk_pixmaps}{$key}});
  #my $pb = $pix->{button} = Gtk::Button->new;
  #$pb->can_default(0);
  #$pb->set_relief('none');
  #$pb->set_border_width(0);
  #$pb->add($pix);
  #$parent->pack_start($pb,@packargs);

  #my $pix = Gtk::Pixmap->new(@{$wg->{gui}{gdk_pixmaps}{$key}});
  #my $pf = $pix->{frame} = Gtk::Frame->new;
  #$pf->set_shadow_type('in');
  #$pf->set_border_width(0);
  #$pf->add($pix);
  #$parent->pack_start($pf,@packargs);

  my $pix  = Gtk::Pixmap->new(@{$wg->{gui}{gdk_pixmaps}{$key}});
  my $pixb = $pix->{box} = Gtk::VBox->new();
  $pixb->set_border_width(2);
  $pixb->add($pix);
  $parent->pack_start($pixb,@packargs);

  return $pix;
}

## $statusbar = $wg->create_statusbar
sub create_statusbar {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $w = $gui->{main};

  my $frame = Gtk::HBox->new(0,0);
  my $sbar  = $frame->{'statusbar'} = Gtk::Statusbar->new;
  $frame->pack_start($sbar,1,1,0);

  my $ibar = $frame->{'iconbar'} = Gtk::Frame->new;
  $ibar->set_shadow_type('in');
  $frame->pack_start($ibar,0,1,0);
  my $ihbox = $ibar->{hbox} = Gtk::HBox->new(1,2);
  $ibar->add($ihbox);
  $frame->{pixmap_on} = $wg->add_statusbar_pixmap('on',$ihbox,0,1,0);
  $frame->{pixmap_up} = $wg->add_statusbar_pixmap('up',$ihbox,0,1,0);
  $frame->{pixmap_attach} = $wg->add_statusbar_pixmap('attach',$ihbox,0,1,0);
  $frame->{pixmap_wep} = $wg->add_statusbar_pixmap('wep',$ihbox,0,1,0);
  $frame->{pixmap_scan} = $wg->add_statusbar_pixmap('noscan',$ihbox,0,1,0);

  ##-- tooltips: argh
  #my $tips = $gui->{tooltips};
  #$tips->set_tip($frame->{pixmap_wep}, 'Current WEP status', 'Tips/WEP');
  #$tips->set_tip($frame->{pixmap_wep}{frame}, 'Current WEP status', '');
  #$tips->set_tip($ibar->{pixmap_wep}{button}, 'Current WEP status', 'Tips/WEP');

  return $frame;
}

##--------------------------------------------------------------
## GUI Initialization: Context Menus
##--------------------------------------------------------------
sub create_context_menus {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $menus = $gui->{ctxmenus} = {};

  my ($cm,$cmi);

  ##--------------------------------------
  ## Menu: Available networks
  $cm = $menus->{available} = Gtk::Menu->new;

  ##-- Available: Connect
  $cmi = $cm->{attach} = Gtk::MenuItem->new("Attach");
  $cmi->signal_connect('activate', sub { $wg->_connect; });
  $cmi->show;
  $cm->append($cmi);

  ##-- Available: Disconnect
  $cmi = $cm->{detach} = Gtk::MenuItem->new("Detach");
  $cmi->signal_connect('activate', sub { $wg->_disconnect; });
  $cmi->show;
  $cm->append($cmi);

  ##-- Available: Separator
  $cmi = $cm->{sep1} = Gtk::MenuItem->new();
  $cmi->show;
  $cm->append($cmi);

  ##-- Available: Edit
  $cmi = $cm->{props} = Gtk::MenuItem->new("Properties");
  $cmi->signal_connect('activate', sub { $wg->available_properties_click; });
  $cmi->show;
  $cm->append($cmi);


  ##--------------------------------------
  ## Menu: Known networks
  $cm = $menus->{known} = Gtk::Menu->new;

  ##-- Known: Add
  $cmi = $cm->{add} = Gtk::MenuItem->new("Add");
  $cmi->signal_connect('activate', sub { $wg->known_add; });
  $cmi->show;
  $cm->append($cmi);

  ##-- Known: Remove
  $cmi = $cm->{remoive} = Gtk::MenuItem->new("Remove");
  $cmi->signal_connect('activate', sub { $wg->known_remove; });
  $cmi->show;
  $cm->append($cmi);

  ##-- Known: Edit
  $cmi = $cm->{edit} = Gtk::MenuItem->new("Edit");
  $cmi->signal_connect('activate', sub { $wg->known_edit; });
  $cmi->show;
  $cm->append($cmi);

  ##-- Known: Separator
  $cmi = $cm->{sep1} = Gtk::MenuItem->new();
  $cmi->show;
  $cm->append($cmi);

  ##-- Known: Set Default
  $cmi = $cm->{default} = Gtk::MenuItem->new("Set Default");
  $cmi->signal_connect('activate', sub { $wg->known_set_default; });
  $cmi->show;
  $cm->append($cmi);

  ##-- Known: Separator #2
  $cmi = $cm->{sep2} = Gtk::MenuItem->new();
  $cmi->show;
  $cm->append($cmi);

  ##-- Known: Revert
  $cmi = $cm->{revert} = Gtk::MenuItem->new("Revert");
  $cmi->signal_connect('activate', sub { $wg->known_refresh(revert=>1); });
  $cmi->show;
  $cm->append($cmi);


  ##-- return
  return $menus;
}

##======================================================================
## Update / Instantiation
##======================================================================

## undef = $wg->update()
##  + %args:
##    force=>$bool
sub update {
  my ($wg,%args) = @_;

  ##-- possibly force recomputation of status flags
  $wg->{status} = $wg->getStatus() if ($args{force});

  $wg->update_toolbar();
  $wg->update_status_icons();
  $wg->update_current_network();
}

##--------------------------------------------------------------
## Update: Toolbar
sub update_toolbar {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $tbar = $gui->{toolbar};
  my $tips = $gui->{tips};

  my $status = $wg->status;

  ##-- On/Off
  if ($status->{ison}) {
    ##-- +on
    $tbar->{on}{pixmap}->set(@{$gui->{gdk_pixmaps}{off}});
    $tbar->{on}{label}->set('Off');
    $tips->set_tip($tbar->{on}, 'Disable WLAN device', $tbar->{on}{tippriv});
  }
  else {
    ##-- -on
    $tbar->{on}{pixmap}->set(@{$gui->{gdk_pixmaps}{on}});
    $tbar->{on}{label}->set('On');
    $tips->set_tip($tbar->{on}, 'Enable WLAN device', $tbar->{on}{tippriv});
  }

  ##-- Up/Down
  if ($status->{isup}) {
    ##-- +up
    $tbar->{up}{pixmap}->set(@{$gui->{gdk_pixmaps}{down}});
    $tbar->{up}{label}->set('Down');
    $tips->set_tip($tbar->{up}, 'Run down WLAN device', $tbar->{up}{tippriv});
  }
  else {
    ##-- -up
    $tbar->{up}{pixmap}->set(@{$gui->{gdk_pixmaps}{up}});
    $tbar->{up}{label}->set('Up');
    $tips->set_tip($tbar->{up}, 'Run up WLAN device', $tbar->{up}{tippriv});
  }

  ##-- Attach/Detach
  if ($status->{isassoc}) {
    ##-- +attached
    $tbar->{attach}{pixmap}->set(@{$gui->{gdk_pixmaps}{detach}});
    $tbar->{attach}{label}->set('Detach');
    $tips->set_tip($tbar->{attach}, 'Detach from current network', $tbar->{attach}{tippriv});
  }
  else {
    ##-- -attached
    $tbar->{attach}{pixmap}->set(@{$gui->{gdk_pixmaps}{attach}});
    $tbar->{attach}{label}->set('Attach');
    $tips->set_tip($tbar->{attach}, 'Attach to selected network', $tbar->{attach}{tippriv});
  }
}


##--------------------------------------------------------------
## Update: Status icons
sub update_status_icons {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $sbar = $gui->{statusbar};

  my ($pix);
  my $status = $wg->status;

  $pix = $sbar->{pixmap_on};
  if ($status->{ison}) {
    ##-- +on
    $sbar->{pixmap_on}->set(@{$gui->{gdk_pixmaps}{on}});

    if ($status->{isup}) {
      ##-- +on, +up
      $sbar->{pixmap_up}->set(@{$gui->{gdk_pixmaps}{up}});
      if ($status->{isassoc}) {
	##-- +on, +up, +assoc
	$sbar->{pixmap_attach}->set(@{$gui->{gdk_pixmaps}{attached}});
      } else {
	##-- +on, +up, -assoc
	$sbar->{pixmap_attach}->set(@{$gui->{gdk_pixmaps}{detached}});
      }
    } else {
      ##-- +on, -up, => -assoc
      $sbar->{pixmap_up}->set(@{$gui->{gdk_pixmaps}{down}});
      $sbar->{pixmap_attach}->set(@{$gui->{gdk_pixmaps}{detached}});
    }
  } else {
    ##-- -on => -up, -assoc
    $sbar->{pixmap_on}->set(@{$gui->{gdk_pixmaps}{off}});
    $sbar->{pixmap_up}->set(@{$gui->{gdk_pixmaps}{down}});
    $sbar->{pixmap_attach}->set(@{$gui->{gdk_pixmaps}{detached}});
  }

  if ($status->{isencrypted}) {
    $sbar->{pixmap_wep}->set(@{$gui->{gdk_pixmaps}{wep}});
  } else {
    $sbar->{pixmap_wep}->set(@{$gui->{gdk_pixmaps}{nowep}});
  }
}


##--------------------------------------------------------------
## Update: Current network
sub update_current_network {
  my $wg = shift;
  my $gui = $wg->{gui};

  ##-- current-network info frame
  my $cframe = $gui->{infobox}{current};
  my $tstatus = $wg->statusText;
  $cframe->{name}[1]->set($tstatus->{name});
  $cframe->{essid}[1]->set($tstatus->{essid});
  $cframe->{ap}[1]->set($tstatus->{ap});
  $cframe->{ipaddr}[1]->set($tstatus->{ipaddr});

  ##-- current network "attached" pixmap (works on attach, not on detach)
  ##-- clear 'attached' pixmaps in clist
  my $status = $wg->status;
  my $curnet = $wg->{current_network};
  my ($row,$ap);
  foreach $row (0..$#{$gui->{row2ap}}) {
    $ap = $gui->{row2ap}[$row];
    if ($status->{isassoc} && $curnet && $curnet->{ap} && $curnet->{ap} eq $ap)
      {
	$gui->{clist}->set_pixmap($row,0,@{$gui->{gdk_pixmaps}{attached_cl}});
      }
    else
      {
	$gui->{clist}->set_pixmap($row,0,@{$gui->{gdk_pixmaps}{detached_cl}});
      }
  }
}

##======================================================================
## Messages
##======================================================================

## undef = $wd->vmsg($level,@message)
sub vmsg_argh {
  my ($wg,$level,@message) = @_;
  $wg->{verbose} = 'default' if (!defined($wg->{verbose}));
  if ($wg->{gui} && $wg->{gui}{statusbar}) {
    if ($vlevels{$level} <= $vlevels{$wg->{verbose}}) {
      my $text = join('',@message);
      $text =~ s/\n*$//m;
      $wg->{gui}{statusbar}{statusbar}->pop(1);
      $wg->{gui}{statusbar}{statusbar}->push(1,$text);
      $wg->{gui}{statusbar}{statusbar}->show_all;
    }
  }
  $wg->SUPER::vmsg($level,@message);
}


##======================================================================
## Callbacks
##======================================================================

##--------------------------------------------------------------
## Callbacks: Wrappers
##--------------------------------------------------------------

sub _wrap_gui {
  my ($dorefresh,@methods) = @_;
  return sub {
    my ($wg,@args) = @_;
    my $gui = $wg->{gui};

    ##-- remove any old timeouts
    if (defined($gui->{wrapper_timeout_id})) {
      Gtk->timeout_remove($gui->{wrapper_timeout_id});
      delete($gui->{wrapper_timeout_id});
    }

    $gui->{wrapper_timeout_id} =
      Gtk->timeout_add(500,
		       sub {
			 my ($method);
			 foreach (@methods) {
			   $method = UNIVERSAL::can($wg,$_);
			   $method->($wg,@args);
			   sleep(1);
			   $wg->update;
			 }
			 if ($dorefresh) {
			   $wg->_refresh();
			   $wg->update;
			 }

			 ##-- delete old wrapper id
			 if (defined($gui->{wrapper_timeout_id})) {
			   Gtk->timeout_remove($gui->{wrapper_timeout_id});
			   delete($gui->{wrapper_timeout_id});
			 }
		       });
  }
}

##--------------------------------------------------------------
## Enable/Disable

*_enable  = _wrap_gui(0,'enable');
*_disable = _wrap_gui(0,qw(disconnect ifdown disable));

##--------------------------------------------------------------
## ifup/ifdown

*_ifup   = _wrap_gui(1,qw(on ifup));
*_ifdown = _wrap_gui(1,qw(disconnect ifdown));

##--------------------------------------------------------------
## Connect/Disconnect
*_connect_wrapper    = _wrap_gui(1,'connect');
sub _connect {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $clist = $gui->{clist};

  my $row = $clist->selection;
  if (defined($row)) {
    return $wg->_connect_wrapper($clist->get_row_data($row));
  }

  $wg->confirm_dialog('confirm_default_connect',
		      ref($wg)."::Default Connection",
		      sub {
			$wg->_connect_wrapper();
		      },
		      "No network selected!\n",
		      "Use default parameters?");


}

*_disconnect_wrapper = *_disconnect = _wrap_gui(0,'disconnect');

##--------------------------------------------------------------
## Callbacks: Toggle (for toolbar)
##--------------------------------------------------------------

## $toggle_sup = _toggle_wrapper($status_key, $off2on_sub, $on2off_sub);
sub _toggle_wrapper {
  my ($key,$onsub,$offsub) = @_;
  return sub {
    my $wg = shift;
    my $status = $wg->status;
    if ($status && $status->{$key}) {
      $offsub->($wg,@_);
    } else {
      $onsub->($wg,@_);
    }
  };
}

*toggle_on     = _toggle_wrapper('ison', \&_enable, \&_disable);
*toggle_up     = _toggle_wrapper('isup', \&_ifup, \&_ifdown);
*toggle_attach = _toggle_wrapper('isassoc', \&_connect, \&_disconnect);

##--------------------------------------------------------------
## Callbacks: Scan/Refresh
##--------------------------------------------------------------

##-- refresh scan list
sub _refresh {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $clist = $gui->{clist};

  ##-- clear the clist
  $clist->clear;

  ##-- rescan
  my @scanned = $wg->scanMatch();
  $gui->{row2ap} = [] if (!$gui->{row2ap}); @{$gui->{row2ap}} = qw();
  $gui->{ap2row} = {} if (!$gui->{ap2row}); %{$gui->{ap2row}} = qw();
  $gui->{ap2qhist} = {} if (!$gui->{ap2qhist}); #%{$gui->{ap2qhist}} = qw();

  ##-- current network
  my $curap = '';
  $curap = $wg->{current_network}{ap}
    if ($wg->{current_network} && $wg->{current_network}{ap});

  ##-- link-quality pixmap preparations
  my $cmap =  $gui->{colormap};
  my $bg   = $clist->style->white;
  my $lmax = $gui->{link_max};
  my $src = $gui->{gdk_pixmaps}{link}[0];
  my @size = ($src->get_size)[1,0];

  ##-- update scan-list
  my ($row,$scan,$ap,$net,$iwparams,$qfrac,$gc,$pix,$msk, $qhist);
  my $status = $wg->status;
  foreach $row (0..$#scanned) {
    $net                 = $scanned[$row];
    $scan                = $net->{scan};
    $iwparams            = $net->{iwparams};
    $ap                  = $net->{ap};
    $gui->{row2ap}[$row] = $ap;
    $gui->{ap2row}{$ap}  = $row;

    ##-- text content
    $clist->append('',                                        ##-- status
		   ($net->{name}  ? $net->{name} : '?'),      ##-- name
		   ($iwparams->{essid} ? $iwparams->{essid}   ##-- essid
		    : ($scan->{essid} ? $scan->{essid} : '?')),
		   '',#($scan->{encryption} ? 'yes' : 'no'),  ##-- wep?
		   '',#($scan->{quality} ? $scan->{quality} : 0) ##-- link quality
		  );

    ##-- data: net
    $clist->set_row_data($row,$net);

    ##-- known/unknown
    $clist->set_foreground($row, ($net && $net->{name} !~ /^unknown/
				  ? $gui->{colors}{known}
				  : $gui->{colors}{unknown}));

    ##-- wep pixmap
    if (defined($scan->{key}) && $scan->{key} ne 'off') {
      $clist->set_pixmap($row,3, @{$gui->{gdk_pixmaps}{wep_small}});
    }
    #else { $clist->set_pixmap($row,2, @{$gui->{gdk_pixmaps}{nowep}}); }

    ##-- current+attached ?
    if ($ap eq $curap) {
      $clist->set_pixmap($row,0, @{$gui->{gdk_pixmaps}{attached_cl}});
    }

    ##-- Link Quality
    if ($gui->{link_nsamples} > 1 &&  $ap =~ /[1-9a-fA-F]/) {
      $qhist = $gui->{ap2qhist}{$ap};
      $qhist = $gui->{ap2qhist}{$ap} = [] if (!defined($gui->{ap2qhist}{$ap}));

      push(@$qhist, $scan->{link_qual});
      shift(@$qhist) while (@$qhist > $gui->{link_nsamples});
      $qfrac = 0;
      $qfrac += $_/$gui->{link_max} foreach (@$qhist);
      $qfrac /= @$qhist;
      $qfrac = 1.0 if ($qfrac > 1.0);
    } else {
      $qfrac = $scan->{link_qual}/$gui->{link_max};
    }

    ##-- Current?
    #if ($status->{isassoc} && $net->{ap} eq $wg->{current_network}->{ap}) {
    #  $gui->{clist}->set_pixmap($row,0,@{$gui->{gdk_pixmaps}{attached_cl}});
    #}

    ##-- Link: Pixmap
    #$qfrac = 1.0;
    ($pix,$msk) = Gtk::Gdk::Pixmap->new($gui->{main}->window, @size);
    $gc = Gtk::Gdk::GC->new($pix);

    ##-- Link: background (white rect)
    $gc->set_foreground($bg);
    #$pix->draw_rectangle($gc, 1, 0,0, $size[0]-1,$size[1]-1);
    $pix->draw_rectangle($gc, 1, 0,0, $size[0]-0,$size[1]-0);

    ##-- Link: foreground (copy)
    $pix->draw_pixmap($gc, $src, 0,0, 1,1, $qfrac*($size[0]-2), $size[1]-2);
    #$pix->draw_pixmap($gc, $src, 0,0, 0,0, $qfrac*($size[0]-0), $size[1]-0);

    ##-- Link: limit (max line)
    #$gc->set_foreground($lm);
    #$pix->draw_line($gc, $qfrac*($size[0]-2), 1, $qfrac*($size[0]-2), $size[1]-2);

    ##-- Link: border frame
    $gc->set_foreground($gui->{colors}{border});
    $pix->draw_rectangle($gc, 0, 0,0, $size[0]-1,$size[1]-1);

    ##-- Link: label text
    $gc->set_foreground($gui->{colors}{qlabel});
    $pix->draw_string($gui->{fonts}{value}, $gc,
		      #$size[0]-25, $size[1]-1,
		      $size[0]/2-12, $size[1]-1,
		      #"100%",
		      sprintf("%d%%", 100*$qfrac),
		      #$scan->{quality}
		     );

    $clist->set_pixmap($row,4,$pix,$msk);
  }

  ##-- re-select network
  $clist->select_row($gui->{ap2row}{$gui->{selected_network}{ap}}, 0)
    if ($gui->{selected_network}
	&& $gui->{selected_network}{ap}
	&& defined($gui->{ap2row}{$gui->{selected_network}{ap}}));
}

##--------------------------------------------------------------
## Callbacks: Autoscan
##--------------------------------------------------------------
sub autoscan_cb {
  my ($wg,%args) = @_;
  my $gui = $wg->{gui};
  my $tbar = $gui->{toolbar};
  my $tips = $gui->{tips};

  $gui->{autoscan} = !$gui->{autoscan} if ($args{toggle});

  ##-- remove any old callbacks
  Gtk->timeout_remove($gui->{autoscan_cbid}) if (defined($gui->{autoscan_cbid}));
  delete($gui->{autoscan_cbid});


  if ($gui->{autoscan} && $wg->status->{ison}) {
    ##-- +autoscan: status icon
    $gui->{statusbar}{pixmap_scan}->set(@{$gui->{gdk_pixmaps}{scan}});
    $wg->vmsg('debug2', "Autoscanning ", ++$gui->{autoscan_n}, "\n");

    ##-- +autoscan: menu entry
    $gui->{menubar}{factory}->get_item('/Actions/Autoscan')->active(1);

    ##-- +autoscan: toolbar button
    $tbar->{autoscan}{pixmap}->set(@{$gui->{gdk_pixmaps}{noscan}});
    $tbar->{autoscan}{label}->set('No Scan');
    $tips->set_tip($tbar->{autoscan},
		   'Stop auto-scanning for available networks',
		   $tbar->{autoscan}{tippriv});

    ##-- +autoscan: refresh, timeout
    $wg->_refresh;
    $gui->{autoscan_cbid} = Gtk->timeout_add($gui->{autoscan_interval},
					     sub { $wg->autoscan_cb(toggle=>0); });
  }
  else {
    ##-- -autoscan: status icon
    $gui->{statusbar}{pixmap_scan}->set(@{$gui->{gdk_pixmaps}{noscan}});

    ##-- +autoscan: menu entry
    $gui->{menubar}{factory}->get_item('/Actions/Autoscan')->active(0);

    ##-- -autoscan: toolbar button
    $tbar->{autoscan}{pixmap}->set(@{$gui->{gdk_pixmaps}{scan}});
    $tbar->{autoscan}{label}->set('Autoscan');
    $tips->set_tip($tbar->{autoscan},
		   'Auto-scan for available networks',
		   $tbar->{autoscan}{tippriv});
  }
}

##--------------------------------------------------------------
## Callbacks: Available Networks: Select/Unselect
##--------------------------------------------------------------

sub clist_select {
  my ($wg,$clist,$row,$col,$event) = @_;
  my $gui = $wg->{gui};

  my $ap    = $gui->{row2ap}[$row];
  my $net      = $wg->{networks_byap}{$ap};
  my $scan     = $net->{scan};
  my $iwparams = $net->{iwparams};

  ##-- Update "selected network" infobox
  my $info = $gui->{infobox}{selected};
  $info->{name}[1]->set($net && $net->{name} ? $net->{name} : '(unknown)');
  $info->{essid}[1]->set($scan->{essid} ? $scan->{essid}
			 : ($iwparams->{essid} ? $iwparams->{essid} : '?'));
  $info->{ap}[1]->set($scan->{ap});

  $gui->{selected_network} = $net;

  ##-- event-dependent signals
  if ($event && $event->{type}) {
    ##-- double-click: connect/disconnect
    if ($event->{type} eq '2button_press') {
      if ($wg->status->{isassoc}) {
	$wg->_disconnect;
      } else {
	$wg->_connect($net);
      }
    }
    #else { print "Unknown event: type=$event->{type}\n";   }
  }
}



sub clist_unselect {
  my ($wg,$clist,$row,$col,$event) = @_;
  my $gui = $wg->{gui};

  ##-- Update "selected network" infobox
  my $info = $gui->{infobox}{selected};
  $info->{name}[1]->set('(none)');
  $info->{essid}[1]->set('off');
  $info->{ap}[1]->set('off');

  delete($gui->{selected_network});
}

##--------------------------------------------------------------
## Callbacks: Available Networks: Button Press (context menu)
##--------------------------------------------------------------

sub clist_click {
  my ($wg,$clist,$event) = @_;

  ##-- right-click: context menu
  if ($event
      && $event->{type}
      && $event->{type} eq 'button_press'
      && $event->{button} == 3)
    {
      return $wg->error_dialog(undef, 'No available networks!')
	if (!$clist->rows);

      ##-- first, select the nearest row
      my ($x,$y) = @$event{qw(x y)};
      if (defined($x) && defined($y)) {
	my ($row,$col) = $clist->get_selection_info($x,$y);

	return $wg->error_dialog(undef, 'No available network selected!')
	  if (!defined($row) || !defined($col));

	$clist->select_row($row,$col);
      }

      my $gui = $wg->{gui};
      my $row = $clist->selection;
      return $wg->error_dialog(undef, 'No network selected!')
	if (!defined($row)); ##-- should never happen

      my $net = $clist->get_row_data($row);
      my $menu = $gui->{ctxmenus}{available};
      $menu->set_title("Network $net->{name}");
      $menu->popup(undef,undef, $event->{button}, $event->{time}, undef);
    }
  #else { print "Unknown event: type=$event->{type}\n"; }

  return 0;
}

##--------------------------------------------------------------
## Callbacks: Available Networks: Edit Properties
##--------------------------------------------------------------

sub available_properties_click {
  my $wg = shift;
  my $gui = $wg->{gui};

  my $row = $gui->{clist}->selection;
  my $net = $gui->{clist}->get_row_data($row);

  $wg->known_edit(net=>$net,key=>'available_edit');
  return 0;
}

##--------------------------------------------------------------
## Callbacks: Dialogs: Generic
##--------------------------------------------------------------

## undef = $wg->destroy_dialog(@window_refs_and_gui_keys)
sub destroy_dialog {
  my $wg = shift;
  my ($ref,$key);
  foreach (@_) {
    next if (!defined($_));
    if (ref($_)) { ($ref,$key) = ($_,undef); }
    else         { ($ref,$key) = (\$wg->{gui}{$_},$_); }

    #print "destroy: ref=$ref ; key=", (defined($key) ? $key : '(undef)'), "\n";

    delete($wg->{gui}{$key}) if (defined($key));
    %{$$ref} = qw() if (ref($$ref) && UNIVERSAL::isa($$ref,'HASH'));
    undef($$ref) if (defined($$ref));
  }
  0;
}

##--------------------------------------------------------------
## Callback: Generic Dialog
##--------------------------------------------------------------

## $dlg = $wg->generic_dialog($key,%args)
##  + $key = $gui_key,
##  + %args:
##    title => $title,
##    ref => $scalar_reference,
##    class => $dialog_class,
##    newargs=>$args_to_new()
sub generic_dialog {
  my ($wg,$key,%args) = @_;
  my $gui = $wg->{gui};

  if (!defined($gui->{$key})) {
    $args{class} = 'Gtk::Dialog' if (!$args{class});
    $gui->{$key} = $args{class}->new($args{newargs} ? @{$args{newargs}} : qw());
    $gui->{$key}->signal_connect("destroy",
				 sub { $wg->destroy_dialog($args{ref},$key); });
    $gui->{$key}->signal_connect("delete_event",
				 sub { $wg->destroy_dialog($args{ref},$key); });
    $gui->{$key}->set_title($args{title} ? $args{title} : $args{class});
  }
  #if (!$gui->{$key}->visible) {
  #  $gui->{$key}->show;
  #} else {
  #  $gui->{$key}->destroy;
  #}

  ${$args{ref}} = $gui->{$key} if (defined($args{ref}) && ref($args{ref}));
  return $gui->{$key};
}

##--------------------------------------------------------------
## Callback: Error Dialog
##--------------------------------------------------------------

## $wg->error_dialog($title,@message)
sub error_dialog {
  my ($wg,$title,@message) = @_;
  my $gui = $wg->{gui};

  my ($dlg);
  $wg->generic_dialog('error_dialog', ref=>\$dlg,
		      title=>($title ? $title : (ref($wg).'::Error'))
		     );
  $dlg->set_default_size(300,100);
  $dlg->border_width(5);
  $dlg->show;

  my ($b);
  $b = $dlg->{dismiss} = Gtk::Button->new("Dismiss");
  $b->signal_connect("clicked", sub { $dlg->destroy });
  $b->can_default(1);
  $b->show;
  $dlg->action_area->add($b);

  ##-- hbox
  my $hbox = $dlg->{hbox} = Gtk::HBox->new(0,5);
  $hbox->show;
  $dlg->vbox->pack_start($hbox,1,1,5);

  ##-- pixmap
  my $pix = $dlg->{pix} = Gtk::Pixmap->new(@{$gui->{gdk_pixmaps}{error}});
  $pix->show;
  $hbox->pack_start($pix,0,0,5);

  ##-- subbox
  my $labbox = $dlg->{labbox} = Gtk::HBox->new(0,5);
  $labbox->show;
  $hbox->pack_start($labbox, 1,1,5);

  ##-- error message
  my $lab = Gtk::Label->new(join('',@message));
  $lab->set_line_wrap(1);
  $lab->show;
  $labbox->pack_start($lab, 1,1,5);

  ##-- select the default button
  $dlg->{dismiss}->signal_emit('focus',1);
  $dlg->show_all;

  return 0;
}

## $wg->confirm_dialog($key,$title,$callback,@message)
sub confirm_dialog {
  my ($wg,$key,$title,$callback,@message) = @_;
  my $gui = $wg->{gui};

  my ($dlg);
  $wg->generic_dialog($key,
		      ref=>\$dlg,
		      title=>($title ? $title : (ref($wg).'::Confirm'))
		     );
  $dlg->set_default_size(300,100);
  $dlg->border_width(5);
  $dlg->show;

  my ($b);
  ##-- button: cancel
  $b = $dlg->{cancel} = Gtk::Button->new("Cancel");
  $b->signal_connect("clicked", sub { $dlg->destroy });
  $b->can_default(1);
  $b->show;
  $dlg->action_area->add($b);

  ##-- button: ok
  $b = $dlg->{ok} = Gtk::Button->new("OK");
  $b->signal_connect("clicked", sub { &$callback(); $dlg->destroy; });
  $b->can_default(1);
  $b->show;
  $dlg->action_area->add($b);

  ##-- hbox
  my $hbox = $dlg->{hbox} = Gtk::HBox->new(0,5);
  $hbox->show;
  $dlg->vbox->pack_start($hbox,1,1,5);

  ##-- pixmap
  my $pix = $dlg->{pix} = Gtk::Pixmap->new(@{$gui->{gdk_pixmaps}{question}});
  $pix->show;
  $hbox->pack_start($pix,0,0,5);

  ##-- subbox
  my $labbox = $dlg->{labbox} = Gtk::HBox->new(0,5);
  $labbox->show;
  $hbox->pack_start($labbox, 1,1,5);

  ##-- error message
  my $lab = Gtk::Label->new(join('',@message));
  $lab->set_line_wrap(1);
  $lab->show;
  $labbox->pack_start($lab, 1,1,5);

  ##-- select the default button
  $dlg->{ok}->signal_emit('focus',1);
  $dlg->show_all;
}

##--------------------------------------------------------------
## Callback: About Dialog
##--------------------------------------------------------------

sub display_about {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $prog = $wg->{gui}{name};

  my ($aw);
  $wg->generic_dialog('about', title=>($prog.'::About'), ref=>\$aw);
  $aw->set_default_size(300,200);
  $aw->set_title($prog."::About");
  $aw->border_width(0);
  $aw->show;

  my $btn = $aw->{dismiss} = Gtk::Button->new("Dismiss");
  $btn->signal_connect("clicked", sub { $aw->destroy });
  $btn->can_default(1);
  $aw->action_area->pack_start($btn, 1, 1, 0);
  $btn->show;

  ##-- logo
  my $logo = Gtk::Pixmap->new(@{$gui->{gdk_pixmaps}{logo}});
  $aw->vbox->pack_start($logo,0,0,10);

  ##-- banner
  my $banner = Gtk::Label->new($prog);
  $banner->set_style($gui->{styles}{banner});
  $aw->vbox->pack_start($banner,0,0,5);

  ##-- separator
  my $sep = Gtk::HSeparator->new;
  $aw->vbox->pack_start($sep,0,0,10);

  ##-- table
  my $nrows = 4;
  my $ncols = 2;
  my $tabhb = Gtk::HBox->new;
  my $tab = Gtk::Table->new($nrows,$ncols,0);
  my @attachargs = ([qw(fill expand)], [qw(fill expand)], 5, 5);
  $aw->vbox->pack_start($tabhb,1,1,10);
  $tabhb->pack_start($tab,1,1,25);
  $tab->show;

  $wg->label_value_pair($tab, 0, "Version:", $VERSION);
  $wg->label_value_pair($tab, 1, "Author:", 'Bryan Jurish');
  $wg->label_value_pair($tab, 2, "E-Mail:", '<moocow@ling.uni-potsdam.de>');
  $wg->label_value_pair($tab, 3, "License:", 'GPL');

  ##-- select the button
  $aw->{dismiss}->signal_emit('focus',1);

  $aw->show_all;
}



##--------------------------------------------------------------
## Callback: Preferences dialog
##--------------------------------------------------------------

sub edit_preferences {
  my $wg = shift;
  my $gui = $wg->{gui};

  my ($dlg);
  $wg->generic_dialog('prefs', title=>(ref($wg).'::Preferences'), ref=>\$dlg);
  $dlg->set_default_size(300,200);
  $dlg->border_width(5);
  $dlg->show;

  my $bbox = $dlg->{bbox} = Gtk::HButtonBox->new;
  $bbox->can_default(1);
  $bbox->set_layout_default('end');
  $bbox->show;
  $dlg->action_area->add($bbox);

  my ($b);
  ##-- button: ok
  $b = $dlg->{ok} = Gtk::Button->new("OK");
  $b->signal_connect("clicked", sub { $wg->preferences_apply; $dlg->destroy; });
  $b->can_default(1);
  $b->show;
  $bbox->add($b);

  ##-- button: apply
  $b = $dlg->{apply} = Gtk::Button->new("Apply");
  $b->signal_connect("clicked", sub { $wg->preferences_apply; });
  $b->can_default(1);
  $b->show;
  $bbox->add($b);

  ##-- button: cancel
  $b = $dlg->{cancel} = Gtk::Button->new("Cancel");
  $b->signal_connect("clicked", sub { $dlg->destroy });
  $b->can_default(1);
  $b->show;
  $bbox->add($b);

  ##--------------------------------------
  ## main vbox

  ##-- create table
  my ($nrows,$ncols) = (2,5);
  my $row = 0;
  my $tab = $dlg->{table} = Gtk::Table->new($nrows,$ncols, 0);
  $tab->show;
  $dlg->vbox->pack_start($tab, 1,1,0);

  ##-- device
  my ($widget);
  $widget = Gtk::Combo->new;
  my @devices = $wg->devices;
  $widget->set_popdown_strings(@devices) if (@devices);
  $widget->entry->set_text($wg->{device});
  $dlg->{device} = [$wg->labelled_widget($tab,$row++,
					 'Wireless Device:',$widget,
					 labstyle=>'value')];

  ##-- Module
  $widget = Gtk::Entry->new;
#  $widget->signal_connect('focus-in-event',
#			  sub { $_[0]->select_region(0, length($_[0]->get_text)); });
  $widget->set_text($wg->{module}) if ($wg->{module});
  $dlg->{module} = [$wg->labelled_widget($tab,$row++,
					 'Kernel Module:',$widget,
					 labstyle=>'value')];

  ##-- PCMCIA
  $widget = Gtk::ToggleButton->new;
  $widget->active($wg->{pcmcia} ? 1 : 0);
  my $wlabel = $widget->{label} = Gtk::Label->new($wg->{pcmcia} ? 'Yes' : 'No');
  $wlabel->show;
  $widget->add($wlabel);
  $widget->signal_connect('clicked',
			  sub {
			    $_[0]->{label}->set($_[0]->active ? 'Yes' : 'No');
			  });
  $dlg->{pcmcia} = [$wg->labelled_widget($tab,$row++,
					 'Use PCMCIA:',$widget,
					 labstyle=>'value')];

  ##-- Auto-scan interval
  $widget = Gtk::Entry->new;
  $widget->set_text($gui->{autoscan_interval});
  $dlg->{autoscan_interval} = [$wg->labelled_widget($tab,$row++,
						    'Autoscan Interval (ms):',$widget,
						    labstyle=>'value')];

  ##-- maximum link quality
  $widget = Gtk::Entry->new;
  $widget->set_text($gui->{link_max});
  $dlg->{link_max} = [$wg->labelled_widget($tab,$row++,
					   'Maximum Link Quality:',$widget,
					   labstyle=>'value')];

  ##-- number of link samples
  $widget = Gtk::Entry->new;
  $widget->set_text($gui->{link_nsamples});
  $dlg->{link_nsamples} = [$wg->labelled_widget($tab,$row++,
						'Number of Link Samples:',$widget,
						labstyle=>'value')];


  ##-- select the 'Cancel' button
  $dlg->{cancel}->signal_emit('focus',1);
  $dlg->show_all;
}

##------------------------------------------------------
## Callback: Preferences: Apply
sub preferences_apply {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $dlg = $gui->{prefs};
  return if (!$dlg);

  ##-- device
  $wg->{device} = $dlg->{device}[1]->entry->get_text;

  ##-- module
  $wg->{module} = $dlg->{module}[1]->get_text;

  ##-- pcmcia?
  $wg->{pcmcia} = $dlg->{pcmcia}[1]->active;

  ##-- Auto-scan interval
  $gui->{autoscan_interval} = $dlg->{autoscan_interval}[1]->get_text;

  ##-- maximum link quality
  $gui->{link_max} = $dlg->{link_max}[1]->get_text;

  ##-- number of link samples
  $gui->{link_nsamples} = $dlg->{link_nsamples}[1]->get_text;
}


##--------------------------------------------------------------
## Callback: Known Networks Dialog
##--------------------------------------------------------------

sub edit_known_networks {
  my $wg = shift;
  my $gui = $wg->{gui};

  my ($dlg);
  $wg->generic_dialog('known', title=>(ref($wg).'::Networks'), ref=>\$dlg);
  $dlg->border_width(5);
  $dlg->set_default_size(550,300);
  $dlg->show;

  my $bbox = $dlg->{bbox} = Gtk::HButtonBox->new;
  $bbox->set_layout_default('end');
  $bbox->show;
  $dlg->action_area->add($bbox);

  my ($b);
  $b = $dlg->{ok} = Gtk::Button->new("OK");
  $b->signal_connect("clicked", sub { $wg->known_apply; $dlg->destroy; });
  $b->can_default(1);
  $b->show;
  $bbox->add($b);

  $b = $dlg->{apply} = Gtk::Button->new("Apply");
  $b->signal_connect("clicked", sub { $wg->known_apply; });
  $b->can_default(1);
  $b->show;
  $bbox->add($b);

  $b = $dlg->{cancel} = Gtk::Button->new("Cancel");
  $b->signal_connect("clicked", sub { $dlg->destroy });
  $b->can_default(1);
  $b->show;
  $bbox->add($b);

  ##-- vbox sub-box
  my $hbox = $dlg->{hbox} = Gtk::HBox->new(0,5);
  $hbox->show;
  $dlg->vbox->pack_start($hbox,1,1,5);

  ##-- known network list
  my @titles = ('', 'Name', 'ESSID');
  my $klist = $dlg->{klist} = Gtk::CList->new_with_titles(@titles);
  $klist->set_selection_mode('single');
  $klist->set_column_justification(0,'right');
  $klist->set_column_width(0,20);
  $klist->set_column_width(1,128);
  $klist->set_column_width(2,200);
  $klist->reorderable(1);
  $klist->signal_connect('select_row', sub { $wg->known_select(@_); });
  $klist->signal_connect('button-press-event', sub { $wg->klist_click(@_); });
  $klist->show;

  ##-- ... make it scrollable
  my $scrolled_win = $klist->{scroller} = Gtk::ScrolledWindow->new(undef, undef);
  $scrolled_win->set_policy('automatic', 'automatic');
  $scrolled_win->add($klist);
  $scrolled_win->show;
  $hbox->pack_start($scrolled_win, 1,1,5);

  ##-- add operations buttons
  $bbox = $dlg->{bbox2} = Gtk::VButtonBox->new();
  $bbox->set_layout_default('start');
  $bbox->show;
  $hbox->pack_start($bbox, 0,0,5);

  ##-- buttons: revert
  $b = $dlg->{revert} = Gtk::Button->new('Revert');
  $b->signal_connect('clicked', sub { $wg->known_refresh(revert=>1); });
  $b->show;
  $bbox->add($b);

  ##-- buttons: add
  $b = $dlg->{add} = Gtk::Button->new('Add');
  $b->signal_connect('clicked', sub { $wg->known_add; });
  $b->show;
  $bbox->add($b);

  ##-- buttons: remove
  $b = $dlg->{remove} = Gtk::Button->new('Remove');
  $b->signal_connect('clicked', sub { $wg->known_remove; });
  $b->show;
  $bbox->add($b);

  ##-- buttons: edit
  $b = $dlg->{edit} = Gtk::Button->new('Edit');
  $b->signal_connect('clicked', sub { $wg->known_edit; });
  $b->show;
  $bbox->add($b);

  ##-- buttons: move up
  $b = $dlg->{moveup} = Gtk::Button->new('Move Up');
  $b->signal_connect('clicked', sub { $wg->known_move(-1); });
  $b->show;
  $bbox->add($b);

  ##-- buttons: move down
  $b = $dlg->{movedown} = Gtk::Button->new('Move Down');
  $b->signal_connect('clicked', sub { $wg->known_move(1); });
  $b->show;
  $bbox->add($b);

  ##-- buttons: set default
  $b = $dlg->{default} = Gtk::Button->new('Set Default');
  $b->signal_connect('clicked', sub { $wg->known_set_default; });
  $b->show;
  $bbox->add($b);


  ##-- select the default button
  $dlg->{cancel}->signal_emit('focus',1);
  $dlg->show_all;

  ##-- temporary network editing workspace
  $wg->known_refresh(revert=>1);
}

##------------------------------------------------------
## Callback: Networks: Apply
sub known_apply {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $dlg = $gui->{known};
  #return $wg->error_dialog(undef,"known_apply() not implemented!");

  my $klist = $dlg->{klist};
  $wg->{networks} = Storable::dclone([
				      map {
					$klist->get_row_data($_)
				      } (0..($dlg->{klist}->rows-1))
				     ]);

  $wg->_index;

  ##-- update GUI
  $wg->update(force=>1);
  $wg->_refresh;
}

##------------------------------------------------------
## Callback: Networks: Set Default
sub known_set_default {
  my $wg = shift;
  my $dlg = $wg->{gui}{known};
  my $klist = $dlg->{klist};

  my $row = $klist->selection;
  return $wg->error_dialog(undef, 'No network selected!') if (!defined($row));

  my ($oldrow);
  if ($dlg->{default_network}) {
    $oldrow = $klist->find_row_from_data($dlg->{default_network});
    $klist->get_row_data($oldrow)->{default} = 0 if (defined($oldrow));
  }
  my $net = $klist->get_row_data($row);
  $net->{default} = 1;
  $dlg->{default_network} = $net;

  $wg->known_refresh();
}


##------------------------------------------------------
## Callback: Networks: Select
sub known_select {
  my ($wg,$klist,$row,$col,$event) = @_;

  ##-- edit on double-click
  if ($event && $event->{type} eq '2button_press') {
    $wg->known_edit();
  }
}

##------------------------------------------------------
## Callback: Networks: Button Press (context menu)
sub klist_click {
  my ($wg,$klist,$event) = @_;

  ##-- right-click: context menu
  if ($event
      && $event->{type}
      && $event->{type} eq 'button_press'
      && $event->{button} == 3)
    {
      return $wg->error_dialog(undef, 'No known networks!')
	if (!$klist->rows);

      ##-- first, select the nearest row
      my ($x,$y) = @$event{qw(x y)};
      if (defined($x) && defined($y)) {
	my ($row,$col) = $klist->get_selection_info($x,$y);

	return $wg->error_dialog(undef, 'No known network selected!')
	  if (!defined($row) || !defined($col));

	$klist->select_row($row,$col);
      }

      my $gui = $wg->{gui};
      my $row = $klist->selection;
      return $wg->error_dialog(undef, 'No network selected!')
	if (!defined($row)); ##-- should never happen

      my $net = $klist->get_row_data($row);
      my $menu = $gui->{ctxmenus}{known};
      $menu->set_title("Network $net->{name}");
      $menu->popup(undef,undef, $event->{button}, $event->{time}, undef);
    }
  #else { print "Unknown event: type=$event->{type}\n"; }

  return 0;
}


##------------------------------------------------------
## Callback: Networks: Add
sub known_add {
  my $wg = shift;
  my $dlg = $wg->{gui}{known};
  my $klist = $dlg->{klist};

  my $net = { %{$wg->{iwparams} }, dhcp=>1, };

  $net->{name} = "network".(++$wg->{newctr});
  $net->{name} =~ s/HASH\((0x[\da-fA-F]+)\)/$1/;
  push(@{$dlg->{networks}}, $net);

  $wg->known_edit(net=>$net);
}

##------------------------------------------------------
## Callback: Networks: Remove
sub known_remove {
  my $wg = shift;
  my $gui = $wg->{gui};
  my $dlg = $gui->{known};
  my $klist = $dlg->{klist};

  my $row = $klist->selection();
  if (!defined($row)) {
    $wg->error_dialog(undef,"No network selected!");
    return;
  }
  my $net = $dlg->{row2net}[$row];
  splice(@{$dlg->{networks}}, $row, 1, qw());

  $wg->known_refresh;
}

##------------------------------------------------------
## Callback: Networks: Move
sub known_move {
  my ($wg,$how) = @_;
  my $klist = $wg->{gui}{known}{klist};
  my $from = $klist->selection;

  return $wg->error_dialog(undef,'No network selected!')
    if (!defined($from));

  $klist->row_move($from,$from+$how);
}

##------------------------------------------------------
## Callback: Networks: Refresh
##  + %args: revert=>$bool
sub known_refresh {
  my ($wg,%args) = @_;
  my $gui = $wg->{gui};
  my $dlg = $gui->{known};
  my $klist = $dlg->{klist};

  $dlg->{networks} = Storable::dclone($wg->{networks}) if ($args{revert});
  $klist->clear;

  ##-- add all known nets
  my ($net,$row, $saw_default);
  foreach $row (0..$#{$dlg->{networks}}) {
    $net = $dlg->{networks}[$row];

    $klist->append('',
		   $net->{name},
		   ($net->{iwparams}{essid}
		    ? $net->{iwparams}{essid}
		    : ($net->{iwparams}{essid}='off')),
		  );

    ##-- data (network hash)
    $klist->set_row_data($row, $net);

    ##-- default pixmap
    if ($net->{default} && !$saw_default) {
      $saw_default=1;
      $dlg->{default_network} = $net;
      $klist->set_pixmap($row,0, @{$gui->{gdk_pixmaps}{default}});
    } else {
      $net->{default} = 0;
    }

    ##-- known/unknown
    $klist->set_foreground($row, ($net->{name} && $net->{name} !~ /^unknown/
				  ? $gui->{colors}{known}
				  : $gui->{colors}{unknown}));

  }
}

##------------------------------------------------------
## Callback: Networks: Edit

## undef = $wg->known_edit()
## undef = $wg->known_edit(net=>$net,key=>$dialog_key)
sub known_edit {
  my ($wg,%args) = @_;
  my $net = $args{net};
  my $key = $args{key} ? $args{key} : 'known_edit';
  my $gui = $wg->{gui};

  #$wg->error_dialog(undef, 'known_edit() not yet implemented!');

  ##-- selected network
  if (!$net && $key eq 'known_edit') {
    my $krow = $gui->{known}{klist}->selection;
    return $wg->error_dialog(undef, 'No network selected!') if (!defined($krow));

    $net = $gui->{known}{klist}->get_row_data($krow);
  }

  my ($dlg);
  $wg->generic_dialog($key, ref=>\$dlg,
		      title=>(ref($wg).'::Networks Properties'));
  $dlg->set_default_size(400,250);
  $dlg->border_width(5);
  $dlg->show;

  ##-- buttons
  my $bbox = $dlg->{bbox} = Gtk::HButtonBox->new();
  $bbox->set_layout_default('end');
  $bbox->show;
  $dlg->action_area->add($bbox);

  my ($b);
  ##-- buttons: ok
  $b = $dlg->{ok} = Gtk::Button->new("OK");
  $b->signal_connect("clicked",
		     sub {
		       $wg->known_edit_apply(key=>$key);
		       $dlg->destroy;
		     });
  $b->can_default(1);
  $b->show;
  $bbox->add($b);

  ##-- buttons: apply
  $b = $dlg->{apply} = Gtk::Button->new("Apply");
  $b->signal_connect("clicked", sub { $wg->known_edit_apply(key=>$key); });
  $b->can_default(1);
  $b->show;
  $bbox->add($b);

  ##-- buttons: cancel
  $b = $dlg->{cancel} = Gtk::Button->new("Cancel");
  $b->signal_connect("clicked", sub { $dlg->destroy });
  $b->can_default(1);
  $b->show;
  $bbox->add($b);

  ##---------------------------------
  ## vbox

  ##-- selected network
  $dlg->{net} = $net;

  ##-- Table
  my ($nrows,$ncols) = (5,2);
  my $row = 0;
  my $tab = $dlg->{table} = Gtk::Table->new($nrows,$ncols, 0);
  $tab->show;
  $dlg->vbox->pack_start($tab, 1,1,0);

  my ($widget);

  ##-- Name
  $widget = Gtk::Entry->new;
  $widget->set_text($net->{name});
  #$widget->select_region(0,length($widget->get_text));
  $dlg->{name} = [$wg->labelled_widget($tab,$row++,
				       'Name:',$widget,
				       labstyle=>'value')];

  ##-- ESSID
  $widget = Gtk::Entry->new;
  $widget->set_text($net->{iwparams}{essid});
  $dlg->{essid} = [$wg->labelled_widget($tab,$row++,
					'ESSID:',$widget,
					labstyle=>'value')];

  ##-- AP MAC
  $widget = Gtk::Entry->new;
  $widget->set_text($net->{ap} ? $net->{ap} : 'any');
  $dlg->{ap} = [$wg->labelled_widget($tab,$row++,
					'AP MAC:',$widget,
					labstyle=>'value')];

  ##-- Key
  $widget = Gtk::Entry->new;
  $widget->set_text(defined($net->{iwparams}{key}) ? $net->{iwparams}{key} : 'off');
  $dlg->{key} = [$wg->labelled_widget($tab,$row++,
				      'WEP Key:',$widget,
				      labstyle=>'value')];

  ##-- DHCP?
  #$widget = Gtk::CheckButton->new('');
  $widget = Gtk::ToggleButton->new;
  $widget->{label} = Gtk::Label->new($net->{dhcp} ? 'Yes' : 'No');
  $widget->{label}->show;
  $widget->add($widget->{label});
  $widget->active($net->{dhcp} ? 1 : 0);
  $widget->signal_connect('clicked',
			  sub {
			    $_[0]{label}->set($_[0]->active ? 'Yes' : 'No');
			  });
  $dlg->{dhcp} = [$wg->labelled_widget($tab,$row++,
				       'Use DHCP:',$widget,
				       labstyle=>'value')];


  ##-- default button
  $dlg->{cancel}->signal_emit('focus',1);
  $dlg->show_all;
}

##------------------------------------------------------
## Callback: Networks: Edit: Apply
sub known_edit_apply {
  my ($wg,%args) = @_;
  my $key = $args{key} ? $args{key} : 'known_edit';
  my $dlg = $wg->{gui}{$key};
  my $net = $dlg->{net};

  #$wg->error_dialog(undef,'known_edit_apply() not yet implemented!');

  ##-- Name
  $net->{name} = $dlg->{name}[1]->get_text;

  ##-- ESSID
  my $essid = $dlg->{essid}[1]->get_text;
  if ($essid && $essid ne 'off') {
    $net->{iwparams}{essid} = $essid;
  } else {
    $net->{iwparams}{essid} = 'off';
  }

  ##-- AP MAC
  my $ap = $dlg->{ap}[1]->get_text;
  $ap =~ tr/A-F/a-f/;
  if ($ap) { ##  && $ap ne 'off' && $ap ne 'any'
    $net->{ap} = $ap;
  } else {
    $net->{ap} = 'off';
  }

  ##-- Key
  my $wepkey = $dlg->{key}[1]->get_text;
  if ($wepkey) {  #&& $wepkey ne 'off' && $wepkey ne 'open'
    $net->{iwparams}{key} = $wepkey;
  } else {
    $net->{iwparams}{key} = 'off';
  }

  ##-- DHCP?
  $net->{dhcp} = $dlg->{dhcp}[1]->active;

  ##-- re-generate iwparams (obsolete)
  #$net->{iwparams} = [ %{$net->{iwhash}} ];

  if (defined($wg->{gui}{known})) {
    $wg->known_refresh;
  } else {
    $wg->update(force=>1);
    $wg->_refresh;
  }
}

##======================================================================
## Callbacks: File
##======================================================================

##------------------------------------------------------
## Callbacks: File: Open

sub file_open {
  my $wg = shift;
  my $gui = $wg->{gui};

  my ($fs);
  $wg->generic_dialog('file_open',
		      ref=>\$fs,
		      class=>'Gtk::FileSelection',
		      title=>(ref($wg)."::Open Configuration File"),
		      newargs=>[''],
		     );

  ##-- properties
  $fs->set_filename($gui->{filename}) if ($gui->{filename});

  ##-- buttons
  my ($file);
  $fs->cancel_button->signal_connect('clicked', sub { $fs->destroy; });
  $fs->ok_button->signal_connect('clicked',
				 sub {
				   eval {
				     local $_ = $wg;
				     my $filename = $fs->get_filename;
				     do "$filename";
				     $wg->error_dialog(undef,$@) if ($@);
				     $wg->{gui}{filename} = $filename;
				   };
				   $wg->_index;
				   $fs->destroy;
				 });

  $fs->show;
}


##------------------------------------------------------
## Callbacks: File: Save

## undef = $wg->file_save(%args)
## + %args:
##    force=>$bool,
sub file_save {
  my ($wg,%args) = @_;
  my $gui = $wg->{gui};
  my $filename = $gui->{filename};
  return $wg->file_saveas() if (!$filename);

  ##-- confirmation dialog
  if (!$args{force} && -f $filename) {
    return $wg->confirm_dialog('file_save_confirm',
			       (ref($wg)."::Overwrite File?"),
			       sub {
				 $wg->file_save(force=>1);
			       },
			       "Overwrite old file\n",
			       "\`$filename\' ?",
			      );

  }

  ##-- guts: error
  #return $wg->error_dialog(undef, "file_save() not yet implemented: \`$filename'");

  my $fh = IO::File->new(">$filename");
  return $wg->error_dialog(undef, "Could not write file: $!: \`$filename\'")
    if (!$fh);

  ##-- save configuration
  $fh->print("## File: $filename\n",
	     "##   + Created by ", ref($wg), "\n",
	     "##   + Caveat Editor!\n",
	     "\n",
	     Data::Dumper->Dump([
				 $wg->{verbose},
				 $wg->{device},
				 $wg->{module},
				 $wg->{pcmcia},
				 $wg->{gui}{autoscan_interval},
				 $wg->{gui}{link_max},
				 $wg->{gui}{link_nsamples},
				 $wg->{iwparams},
				 $wg->{networks},
				 $wg->{default_network},
				],
				[
				 '$_->{verbose}',
				 '$_->{device}',
				 '$_->{module}',
				 '$_->{pcmcia}',
				 '$_->{gui}{autoscan_interval}',
				 '$_->{gui}{link_max}',
				 '$_->{gui}{link_nsamples}',
				 '$_->{iwparams}',
				 '$_->{networks}',
				 '$_->{default_network}',
				]));
  $fh->close;
}

##------------------------------------------------------
## Callbacks: File: Save As
sub file_saveas {
  my $wg = shift;
  my $gui = $wg->{gui};

  my ($fs);
  $wg->generic_dialog('file_save',
		      ref=>\$fs,
		      class=>'Gtk::FileSelection',
		      title=>(ref($wg)."::Save Configuration File"),
		      newargs=>[''],
		     );

  ##-- properties
  $fs->set_filename($gui->{filename}) if ($gui->{filename});

  ##-- buttons
  $fs->cancel_button->signal_connect('clicked',
				     sub { $gui->{filename} = ''; $fs->destroy; });
  $fs->ok_button->signal_connect('clicked',
				 sub {
				   my $filename = $fs->get_filename;
				   if ($filename && ! -d $filename) {
				     $gui->{filename} = $filename;
				     $wg->file_save();
				     $fs->destroy;
				   }
				 });
  $fs->show;
}



1;

##======================================================================
## Docs
=pod

=head1 NAME

Wlanctl::GUI - wireless control utilities GUI

=head1 SYNOPSIS

 use Wlanctl::GUI;

 #...

=cut

##======================================================================
## Description
=pod

=head1 DESCRIPTION

Not yet written.

=cut

##=====================================================================
## Footer
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2004, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1),

=cut
